(in-package "AP5")

(defclass instances-class-mixin ()
     ((instances :initform () :accessor instances))
  (:documentation  "Lets a class record its instances."))

(defclass ic-mixin-object () ())

(lcl:defadvice (clos-sys:common-add-named-class ic-mixin-object)
    (proto name superclasses slots options)
  (if (typep proto 'instances-class-mixin)
     (lcl:advice-continue proto name
	    (if (member 'ic-mixin-object superclasses)
		superclasses
	      (append superclasses '(ic-mixin-object)))
	    slots options)
    (lcl:advice-continue proto name superclasses slots options)))

(defmethod make-instance ((class instances-class-mixin) &rest initargs)
  (declare (ignore initargs))
  (let ((instance (call-next-method)))
    (when (typep instance class) (push instance (instances class)))
    instance))

(defmethod update-instance-for-different-class :after
     ((previous standard-object) (current ic-mixin-object) &rest initargs)
  (declare (ignore initargs))
  (pushnew current (instances (class-of current))))

(defmethod update-instance-for-different-class :after
     ((previous ic-mixin-object) (current standard-object) &rest initargs)
  (declare (ignore initargs))
  (let ((oldclass (class-of previous)))
    (unless (typep current oldclass)
      (setf (instances oldclass)
	  (delete current (instances oldclass) :test #'eq :count 1)))))

(defclass enumerable-class
	  (instances-class-mixin  standard-class)
     ())

#+lucid
(defmethod clos-system:validate-superclass
	   ((x enumerable-class)
	    (y standard-class))
  t)



(defun clos-enumerable (name keys class-name)
  (warn-ignored-keys name keys (:arity :adder :deleter :tester :generator)
    `(:representation individual :arity 1
         :tester
	 (lambda (&rest ignore) (declare (ignore ignore))
		    '(lambda (ignore a1)
		       (declare (ignore ignore))
		       (typep a1 ',class-name)))
	 :generator (simplegenerator (output)
			  (instances (find-class ',class-name))
			  )
	 ,.keys)))


(defmacro define-enumerted-class (class-name superclasses slots &rest options
				     &environment env)
  (when (assoc :metaclass options)
    (error "you can't specify your own metaclass for define-enumerted-class"))
  (dolist (c superclasses)
    (unless (typep (find-class c nil env) 'instances-class-mixin)
      (error "superclasses
    
  `(progn (defclass ,class-name ,superclasses ,slots
	    (:metaclass enumerable-class) ,.options)
	  (defrelation ,class-name :representation
	    (clos-enumerable ,class-name))))


;;(defclass adb-class (standard-class) ())
